import java.sql.Connection;
import java.sql.DriverManager;
import java.sql.ResultSet;
import java.sql.Statement;
import java.sql.PreparedStatement;
import java.sql.SQLException;
import java.util.Hashtable;
import java.io.*;

/* Author: Julia Stoyanovich (jds1@cs.columbia.edu) */

public class MutaGeneSys {

  private static String _connectString;
  private static String _inFileName;
  private static String _outFileName;
  private static String _population = null;
  private static String _technology = null;
  private static double _coeff = 0.0;
  private static char _outputFormat = 'S';

  private static int _reqId;
  private static int _indId;
  private static int _hapId_1;
  private static int _hapId_2;

  private static Connection _conn;
  private static Hashtable _snpHT = null;
  private static Hashtable _alleleHT = null;

  private static Connection getConn() throws SQLException, ClassNotFoundException {
    String [] connectArr = _connectString.split("@");
    String driver = "oracle.jdbc.driver.OracleDriver";
    String url = "jdbc:oracle:thin:@localhost:1521:" + connectArr[1];
    String [] userPass = connectArr[0].split("/");
    String username = userPass[0];
    String password = userPass[1];
    Class.forName(driver);
    return DriverManager.getConnection(url, username, password);
  }

  private static void printUsage() {
    System.out.println("Wrong number of arguments");
    System.out.println("MutaGeneSys [user/pass@SID] [inputFile] [outputFile] [population] [technology] [coeff]");
  }
 
  private static void testConnection() throws SQLException {
    Statement st = _conn.createStatement();
    ResultSet rs = st.executeQuery("select to_char(sysdate, 'YYYY-MM-DD HH24:MI') res from dual");
    while (rs.next()) {
	System.out.println(rs.getString("res"));
    }
    rs.close();
    st.close();
  }

  private static void initSnpHT() throws SQLException {
    _snpHT = new Hashtable();

    Statement st = _conn.createStatement();
    ResultSet rs = st.executeQuery("select distinct snpId from SNPs_Used_B36");
    while (rs.next()) {
	_snpHT.put(rs.getString("snpId"), new Integer(1));
    }
    rs.close();
    st.close();
  }

  private static void initAlleleHT() throws SQLException {
    _alleleHT = new Hashtable();
    _alleleHT.put("A", new Integer(1));
    _alleleHT.put("C", new Integer(1));
    _alleleHT.put("G", new Integer(1));
    _alleleHT.put("T", new Integer(1));
  }

  private static void executeUpdate(String query) throws SQLException {

   Statement st = _conn.createStatement();
   st.executeUpdate(query);
   st.close();
  }

  private static int getIntFromDB(String query) throws SQLException {

   Statement st = _conn.createStatement();
   ResultSet rs = st.executeQuery(query);
   rs.next();
   int ret = rs.getInt(1);
   rs.close();
   st.close();
   return ret;
  }

  private static void generateOutput() throws SQLException, IOException {

    String query = "select Markers.Result_Full_XML(" + _reqId + ") from dual";
  
    PreparedStatement pstmt= _conn.prepareStatement(query);
    ResultSet rs = pstmt.executeQuery();

    BufferedWriter outFP = new BufferedWriter(new FileWriter(_outFileName));

    String clobVal = null; 
    while (rs.next()) {
      clobVal = rs.getString(1);
      outFP.write(clobVal + "\n");
    }
    outFP.close();
  }

  private static void associate() throws SQLException {

    String query = "begin Markers.Associate(" + _reqId + "); end;";
    executeUpdate(query);

  }
 
  private static void logAssociationRequest() throws SQLException {
   
    _reqId = getIntFromDB("select reqId_Seq.nextval from dual");
    String query = "insert into Request (reqId, indId, hapId, popId, techId, minConf, ts) values (" +
                    _reqId + "," + _indId + "," + _hapId_1 + ",";
    if (_population.equalsIgnoreCase("any")) {
      query += "null,";
    } else {
      query += "'" + _population + "',";
    }
    if (_technology.equalsIgnoreCase("any")) {
      query += "null,";
    } else {
      int techId = getIntFromDB("select distinct id from Technology where upper(name)  like '%" + 
                   _technology.toUpperCase() + "%'");
      query += techId + ",";
    }
    query += _coeff + ", sysdate)";

    executeUpdate(query);

  }

  private static void loadQueryDataIntoDB() throws SQLException, IOException {

    // generate indId and two hapIds
    _indId = getIntFromDB("select IndId_Seq.nextval from dual");
    _hapId_1 = getIntFromDB("select HapId_Seq.nextval from dual");
    _hapId_2 = getIntFromDB("select HapId_Seq.nextval from dual");

    String query = "insert into Individual (id, popId) values (" + _indId + ", '" + _population + "')";
    executeUpdate(query);

    query = "insert into Haplotype (hapId, indId, snpId, allele) values (?, ?, ?, ?)";

    PreparedStatement hap1_stmt = _conn.prepareStatement(query);
    PreparedStatement hap2_stmt = _conn.prepareStatement(query);
    hap1_stmt.setInt(1, _hapId_1);
    hap1_stmt.setInt(2, _indId);
    hap2_stmt.setInt(1, _hapId_2);
    hap2_stmt.setInt(2, _indId);

    // open input file for reading
    BufferedReader inFP = new BufferedReader(new FileReader(_inFileName));

    // a line may contain several space-separated snp entries, e.g., "rs1234:AC re2345:T"
    String line;
    while ((line = inFP.readLine()) != null) {
        String snpArr[] = line.trim().split(" ");
        for (int i=0; i<snpArr.length; i++) {
          String currArr[] = snpArr[i].split(":"); // split an SNP entry into snpId and allele(s)
         if (_snpHT.containsKey(currArr[0])) { // is this a useful SNP id?
           if (_alleleHT.containsKey(currArr[1].substring(0,1))) { 
		// this is a valid allele of a useful SNP: store in the database
               hap1_stmt.setString(3, currArr[0]);
               hap1_stmt.setString(4, currArr[1].substring(0,1));
               hap1_stmt.executeUpdate();
           }
           if (currArr[1].length() > 1) {
             if (_alleleHT.containsKey(currArr[1].substring(1,2))) { 
		// this is a valid allele of a useful SNP: store in the database
               hap2_stmt.setString(3, currArr[0]);
               hap2_stmt.setString(4, currArr[1].substring(1,2));
               hap2_stmt.executeUpdate();
             }
           }
          } 
        }
    }
    inFP.close();
    hap1_stmt.close();
    hap2_stmt.close();
  }

  public static void main(String[] args) {

    // parse command-line arguments
    if (args.length < 6) {
	printUsage();
        return;
    }  
    _connectString = args[0].trim();
    _inFileName = args[1].trim();
    _outFileName = args[2].trim();
    _population = args[3].trim().toUpperCase();
    _technology = args[4].trim();
    _coeff = Double.parseDouble(args[5].trim());

    try {
    // open database connection
    _conn = getConn();
    _conn.setAutoCommit(false);

    // init a hash table of potentially useful SNPs: this saves load time
    initSnpHT();
    initAlleleHT();

    // load haloptype data into the database
    loadQueryDataIntoDB();

    // log MutaGeneSys request
    logAssociationRequest();

    // generate disease susceptibility hypotheses
    associate();

    // generate output
    generateOutput();

    _conn.commit();
    _conn.close();
    } catch (SQLException sqle) {
	System.err.println(sqle.toString());
    } catch (IOException ioe) {
	System.err.println(ioe.toString());
    } catch (ClassNotFoundException cnfe) {
	System.err.println(cnfe.toString());
    }
  }
}

