-- 
-- Package MARKERS implements marker to disease association
-- Computational Human Genetics, Fall 2006
-- Author: Julia Stoyanovich (jds1@cs.columbia.edu)
-- 

create or replace package MARKERS
as

-- log an association request with specified parameters
function Log_Request (
  in_indId 		IN 	Request.indId%TYPE,
  in_hapId 		IN 	Request.hapId%TYPE,
  in_popId 		IN 	Request.popId%TYPE,
  in_techId 	IN 	Request.techId%TYPE,
  in_resolution IN 	Request.resolution%TYPE,
  in_minConf    IN  Request.minConf%TYPE
) return number;

-- run association for the request
procedure Associate (
  in_request    IN Request.reqId%TYPE
);

-- output evidence to individual's disease succeptibility
function Succeptible_XML (
  in_reqId     IN Request.reqId%TYPE,
  disease      IN varchar
) return clob;

-- output individual data
function Individual_XML (
  in_indId 		IN  Request.indId%TYPE
) return varchar;

-- output haplotype data for an individual
function Haplotype_XML (
  in_indId    IN Request.indId%TYPE
) return clob;

-- output full association result
function Result_Full_XML (
  in_reqId    IN Request.reqId%TYPE
) return clob;

-- output association result summary
function Result_Summary_XML (
  in_reqId    IN Request.reqId%TYPE
) return clob;

-- output full results in XML, with hyperlinks to OMIM
function Result_Web_XML (
  in_reqId    IN Request.reqId%TYPE
) return clob;

-- output full results in XML, with hyperlinks to OMIM
function Result_Web_XML_2 (
  in_reqId    IN Request.reqId%TYPE
) return clob;

function comp_confidence(a in number, b in number) 
   return number;

function comp_resolution(a in number, b in number) 
   return number;

function comp_population(a in varchar2, b in varchar2) 
   return number;

function comp_technology(a in varchar2, b in varchar2) 
   return number;

function omim_id_to_link(omimId in number)
   return varchar2; 

function gbrowse_locus_to_link(chromosome in varchar2, locus in number)
   return varchar2; 

procedure clean_up;

end; -- package declaration
/

create or replace package body MARKERS
as
function Log_Request (
	in_indId 	IN 	Request.indId%TYPE,
        in_hapId 	IN 	Request.hapId%TYPE,
        in_popId 	IN 	Request.popId%TYPE,
        in_techId 	IN 	Request.techId%TYPE,
        in_resolution 	IN 	Request.resolution%TYPE,
        in_minConf      IN      Request.minConf%TYPE
) return number
is
   PRAGMA AUTONOMOUS_TRANSACTION;
   v_reqId number;
begin
        if (in_indId is null) AND (in_hapId is null) then
          DBMS_OUTPUT.put_line('Must specify at least one of indId and hapId');
          return -1;
        end if;
	select ReqId_Seq.nextval into v_reqId from dual;
        if in_hapId is null then
        	insert into Request (reqId, indId, hapId, popId, techId, resolution, minConf, ts)
                  (select distinct v_reqId, in_indId, h.hapId, in_popId, in_techId, in_resolution, in_minConf, sysdate
                   from   Haplotype h
                   where  h.indId = in_indId);
	else
          	insert into Request (reqId, indId, hapId, popId, techId, resolution, minConf, ts)
                  (select distinct v_reqId, h.indId, in_hapId, in_popId, in_techId, in_resolution, in_minConf, sysdate
                   from   Haplotype h
                   where  h.hapId = in_hapId);
	end if;
        commit;
        return v_reqId;
end;
function comp_confidence(a in number, b in number)
   return number is
 ret number;
begin
    if (a is null) or (a < b) then
      ret := 1;
    elsif a > b then
      ret := -1;
    else
      ret := 0;
   end if;
   return (ret);
end;
function comp_resolution(a in number, b in number)
   return number is
 ret number;
begin
    if (a is null) or (b is null) or (a < b) then
      ret := 1;
    elsif a > b then
      ret := -1;
    else
      ret := 0;
   end if;
   return (ret);
end;
function comp_population(a in varchar2, b in varchar2)
   return number is
 ret number;
begin
    if (a is null) or (a = b) then
      ret := 0;
    else
      ret := -1;
   end if;
   return (ret);
end;
function comp_technology(a in varchar2, b in varchar2)
   return number is
 ret number;
begin
    if (a is null) or (b is null) or (a = b) then
      ret := 0;
    else
      ret := -1;
   end if;
   return (ret);
end;
function omim_id_to_link(omimId in number)
   return varchar2 is
 href varchar2(256);
begin
    href := 'http://www.ncbi.nlm.nih.gov/entrez/dispomim.cgi?id=' || omimId;
    return(href);
end;
function gbrowse_locus_to_link(chromosome in varchar2, locus in number)
   return varchar2 is
 href varchar2(256);
begin
    href := 'http://www.hapmap.org/cgi-perl/gbrowse/hapmap_B36/?ref=' || chromosome || ';start=' || (locus - 1000) || ';stop=' || (locus + 1000);
    return(href);
end;
procedure Associate (in_request IN Request.reqId%TYPE) is
   v_ts date;
   v_query varchar2(256);
begin
        select sysdate into v_ts from dual;

        -- clean up
        delete from Result_Full where reqId = in_request;
        delete from Result_Summary where reqId = in_request;

        -- single association into Result_Full
	insert into Result_Full (reqId, snpId_1, allele_1, causal_snpId, causal_allele,
                	techId, resolution, popId, coeff, omimId, zyg, ts)
	(select r.reqId, h.snpId, h.allele, m.tgtSNPId, m.tgtAllele,
                m.techId, m.resolution, m.popId, m.coeff, m.omimId,
                case count(*) when 2 then 'hom' else 'het' end, v_ts
	 from   Request r, Haplotype h, Single_OMIM_Map m
	 where  h.snpId = m.srcSNPId
	 and    h.allele = m.srcAllele
	 and    r.hapId = h.hapId
	 and    r.reqId = in_request
	 and    comp_population(r.popId, m.popId) = 0 -- same population
	 and    comp_technology(r.techId, m.techId) = 0 -- same technology
	 and    comp_resolution(r.resolution, m.resolution) >= 0 -- same or higher res
	 and    comp_confidence(r.minConf, m.coeff) >= 0 -- same or higher confidence
         group by r.reqId, h.snpId, h.allele, m.tgtSNPId, m.tgtAllele, m.techId,
        	  m.resolution, m.popId, m.coeff, m.omimId
        );

        -- double association into Result_Full
	insert into Result_Full(reqId, snpId_1, snpId_2, allele_1, allele_2, causal_snpId, causal_allele,
                                techId, resolution, popId, coeff, omimId, zyg, ts)
	(select r.reqId, h1.snpId, h2.snpId, h1.allele, h2.allele, m.tgtSNPId, m.tgtAllele,
       		m.techId, m.resolution, m.popId, m.coeff, m.omimId,
 				case count(*) when 4 then 'hom hom'
                                              when 2 then 'hom het'
					      when 1 then 'het het' end, v_ts
	 from   Request r, 
                Haplotype h1, Haplotype h2, Double_OMIM_Map m
	 where  h1.snpId = m.srcSNPId_1
         and    h2.snpId = m.srcSNPId_2
	 and    h1.allele = m.srcAllele_1
	 and    h2.allele = m.srcAllele_2
         and    h1.snpId < h2.snpId
	 and    r.indId = h1.indId
	 and    r.indId = h2.indId
	 and    h1.hapId = h2.hapId
         and    r.reqId = in_request
	 and    comp_population(r.popId, m.popId) = 0 -- same population
	 and    comp_technology(r.techId, m.techId) = 0 -- same technology
	 and    comp_resolution(r.resolution, m.resolution) >= 0 -- same or higher res
	 and    comp_confidence(r.minConf, m.coeff) >= 0 -- same or higher confidence
         group by r.reqId, h1.snpId, h2.snpId, h1.allele, h2.allele, m.techId,
       		m.resolution, m.popId, m.coeff, m.omimId, m.tgtSNPId, m.tgtAllele
        );
        -- finally, generate a summary of the findings
        insert into Result_Summary(reqId, hapId, snpId_1, snpId_2, omimId, title, zyg, ts)
        (select distinct r.reqId, r.hapId, r.snpId_1, r.snpId_2, r.omimId, o.title, r.zyg, r.ts
         from   Result_Full r, Omim o
         where  r.reqId = in_request
         and    r.omimId = o.id);
       commit;
end;

procedure clean_up is
 currId number;
begin
  -- Delete all rows from Haplotype that correspond to requests that are at
  -- least 2 hours old.
  select max(indId) into currId from request where ts < sysdate - (120/1440);
  delete from Haplotype where indId <= currId;
  commit;
end;

function Succeptible_XML (in_reqId Request.reqId%TYPE, disease IN varchar)
return clob is
 PRAGMA AUTONOMOUS_TRANSACTION;
 ret clob;
begin
  delete from stage;
  insert into stage (
        select distinct reqId, title, Markers.omim_id_to_link(omimId) href from Result_Summary
        where  reqId = in_reqId
        and    upper(title) like ('%' || upper(disease) || '%')
  );
  select dbms_xmlgen.getxml('select reqId, title, href from stage order by title')
  into ret from dual;
  commit;
  return ret;
end;
function Individual_XML (in_indId Request.indId%TYPE)
return varchar is
 ret varchar2(1024);
begin
  select dbms_xmlgen.getxml('select * from Individual where id = ' || in_indId)
  into ret from dual;
  return ret;
end;
function Haplotype_XML (in_indId Request.indId%TYPE)
return clob is
 ret clob;
begin
  select dbms_xmlgen.getxml('select * from Haplotype where indId = ' || in_indId)
  into ret from dual;
  return ret;
end;
function Result_Full_XML (in_reqId Request.reqId%TYPE)
return clob is
 ret clob;
begin
  select dbms_xmlgen.getxml('select * from Result_Full where reqId = ' || in_reqId)
  into ret from dual;
  return ret;
end;
function Result_Web_XML (in_reqId Request.reqId%TYPE)
return clob is
 ret clob;
begin
  select dbms_xmlgen.getxml('select r.SNPId_1, r.allele_1, r.SNPId_2, r.allele_2, r.causal_snpId, r.causal_allele, r.zyg, r.popId, r.coeff, o.id, o.title, Markers.omim_id_to_link(o.id) as url from Result_Full r, OMIM o where r.reqId = ' || in_reqId || ' and r.omimId = o.id')
  into ret from dual;
  return ret;
end;

function Result_Web_XML_2 (in_reqId Request.reqId%TYPE)
return clob is
 xml_result clob;
 v1           VARCHAR2(10);
 qrycontext   DBMS_XMLQUERY.ctxHandle;
 query varchar(256);
begin
 v1 := in_reqId;
 qrycontext :=
      DBMS_XMLQUERY.newContext
        ('select r.SNPId_1, r.allele_1, r.SNPId_2, r.allele_2, r.causal_snpId, r.causal_allele, r.zyg, r.popId, r.coeff, o.id, o.title, Markers.omim_id_to_link(o.id) as url from Result_Full r, OMIM o where r.reqId = :1 and r.omimId = o.id');
  DBMS_XMLQUERY.setBindValue (qrycontext, '1', v1);
  xml_result := DBMS_XMLQUERY.getXML (qrycontext);
  return xml_result;
end;

function Result_Summary_XML (in_reqId Request.reqId%TYPE)
return clob is
 ret clob;
begin
  select dbms_xmlgen.getxml('select r.SNPId_1, r.allele_1, r.SNPId_2, r.allele_2, r.causal_snpId, r.causal_allele, r.zyg, r.popId, r.coeff, o.id, o.title, Markers.omim_id_to_link(o.id) as url from Result_Full r, OMIM o where r.omimId = o.id and r.reqId = ' || in_reqId)
  into ret from dual;
  return ret;
end;
end; -- package body
/
